"""Manim's (internal) color data structure and some utilities for
color conversion.

This module contains the implementation of :class:`.ManimColor`,
the data structure internally used to represent colors.

The preferred way of using these colors is by importing their constants from manim:

.. code-block:: pycon

    >>> from manim import RED, GREEN, BLUE
    >>> print(RED)
    #FC6255

Note this way uses the name of the colors in UPPERCASE.

.. note::

    The colors of type "C" have an alias equal to the colorname without a letter,
    e.g. GREEN = GREEN_C
"""

from __future__ import annotations

import colorsys

# logger = _config.logger
import random
import re
from typing import Any, Sequence, TypeVar, Union, overload

import numpy as np
import numpy.typing as npt
from typing_extensions import Self, TypeAlias

from manim.typing import (
    HSV_Array_Float,
    HSV_Tuple_Float,
    ManimColorDType,
    ManimColorInternal,
    RGB_Array_Float,
    RGB_Array_Int,
    RGB_Tuple_Float,
    RGB_Tuple_Int,
    RGBA_Array_Float,
    RGBA_Array_Int,
    RGBA_Tuple_Float,
    RGBA_Tuple_Int,
)

from ...utils.space_ops import normalize

# import manim._config as _config

re_hex = re.compile("((?<=#)|(?<=0x))[A-F0-9]{6,8}", re.IGNORECASE)


class ManimColor:
    """Internal representation of a color.

    The ManimColor class is the main class for the representation of a color.
    It's internal representation is a 4 element array of floats corresponding
    to a [r,g,b,a] value where r,g,b,a can be between 0 to 1.

    This is done in order to reduce the amount of color inconsistencies by constantly
    casting between integers and floats which introduces errors.

    The class can accept any value of type :class:`ParsableManimColor` i.e.

    ManimColor, int, str, RGB_Tuple_Int, RGB_Tuple_Float, RGBA_Tuple_Int, RGBA_Tuple_Float, RGB_Array_Int,
    RGB_Array_Float, RGBA_Array_Int, RGBA_Array_Float

    ManimColor itself only accepts singular values and will directly interpret them into a single color if possible
    Be careful when passing strings to ManimColor it can create a big overhead for the color processing.

    If you want to parse a list of colors use the function :meth:`parse` in :class:`ManimColor` which assumes that
    you are going to pass a list of color so arrays will not be interpreted as a single color.

    .. warning::
        If you pass an array of numbers to :meth:`parse` it will interpret the r,g,b,a numbers in that array as colors
        so instead of the expect singular color you get and array with 4 colors.

    For conversion behaviors see the ``_internal`` functions for further documentation

    You can create a ``ManimColor`` instance via its classmethods. See the respective methods for more info.

    .. code-block:: python

        mycolor = ManimColor.from_rgb((0, 1, 0.4, 0.5))
        myothercolor = ManimColor.from_rgb((153, 255, 255))

    You can also convert between different color spaces:

    .. code-block:: python

        mycolor_hex = mycolor.to_hex()
        myoriginalcolor = ManimColor.from_hex(mycolor_hex).to_hsv()

    Parameters
    ----------
    value
        Some representation of a color (e.g., a string or
        a suitable tuple). The default ``None`` is ``BLACK``.
    alpha
        The opacity of the color. By default, colors are
        fully opaque (value 1.0).
    """

    def __init__(
        self,
        value: ParsableManimColor | None,
        alpha: float = 1.0,
    ) -> None:
        if value is None:
            self._internal_value = np.array((0, 0, 0, alpha), dtype=ManimColorDType)
        elif isinstance(value, ManimColor):
            # logger.info(
            #     "ManimColor was passed another ManimColor. This is probably not what "
            #     "you want. Created a copy of the passed ManimColor instead."
            # )
            self._internal_value = value._internal_value
        elif isinstance(value, int):
            self._internal_value = ManimColor._internal_from_integer(value, alpha)
        elif isinstance(value, str):
            result = re_hex.search(value)
            if result is not None:
                self._internal_value = ManimColor._internal_from_hex_string(
                    result.group(), alpha
                )
            else:
                # This is not expected to be called on module initialization time
                # It can be horribly slow to convert a string to a color because
                # it has to access the dictionary of colors and find the right color
                self._internal_value = ManimColor._internal_from_string(value)
        elif isinstance(value, (list, tuple, np.ndarray)):
            length = len(value)
            if all(isinstance(x, float) for x in value):
                if length == 3:
                    self._internal_value = ManimColor._internal_from_rgb(value, alpha)  # type: ignore
                elif length == 4:
                    self._internal_value = ManimColor._internal_from_rgba(value)  # type: ignore
                else:
                    raise ValueError(
                        f"ManimColor only accepts lists/tuples/arrays of length 3 or 4, not {length}"
                    )
            else:
                if length == 3:
                    self._internal_value = ManimColor._internal_from_int_rgb(
                        value, alpha  # type: ignore
                    )
                elif length == 4:
                    self._internal_value = ManimColor._internal_from_int_rgba(value)  # type: ignore
                else:
                    raise ValueError(
                        f"ManimColor only accepts lists/tuples/arrays of length 3 or 4, not {length}"
                    )
        elif hasattr(value, "get_hex") and callable(value.get_hex):
            result = re_hex.search(value.get_hex())
            if result is None:
                raise ValueError(f"Failed to parse a color from {value}")

            self._internal_value = ManimColor._internal_from_hex_string(
                result.group(), alpha
            )
        else:
            # logger.error(f"Invalid color value: {value}")
            raise TypeError(
                "ManimColor only accepts int, str, list[int, int, int], "
                "list[int, int, int, int], list[float, float, float], "
                f"list[float, float, float, float], not {type(value)}"
            )

    @property
    def _internal_value(self) -> ManimColorInternal:
        """Returns the internal value of the current Manim color [r,g,b,a] float array

        Returns
        -------
        ManimColorInternal
            internal color representation
        """
        return self.__value

    @_internal_value.setter
    def _internal_value(self, value: ManimColorInternal) -> None:
        """Overwrites the internal color value of the ManimColor object

        Parameters
        ----------
        value : ManimColorInternal
            The value which will overwrite the current color

        Raises
        ------
        TypeError
            Raises a TypeError if an invalid array is passed
        """
        if not isinstance(value, np.ndarray):
            raise TypeError("value must be a numpy array")
        if value.shape[0] != 4:
            raise TypeError("Array must have 4 values exactly")
        self.__value: ManimColorInternal = value

    @staticmethod
    def _internal_from_integer(value: int, alpha: float) -> ManimColorInternal:
        return np.asarray(
            (
                ((value >> 16) & 0xFF) / 255,
                ((value >> 8) & 0xFF) / 255,
                ((value >> 0) & 0xFF) / 255,
                alpha,
            ),
            dtype=ManimColorDType,
        )

    # TODO: Maybe make 8 nibble hex also convertible ?
    @staticmethod
    def _internal_from_hex_string(hex: str, alpha: float) -> ManimColorInternal:
        """Internal function for converting a hex string into the internal representation of a ManimColor.

        .. warning::
            This does not accept any prefixes like # or similar in front of the hex string.
            This is just intended for the raw hex part

        *For internal use only*

        Parameters
        ----------
        hex : str
            hex string to be parsed
        alpha : float
            alpha value used for the color

        Returns
        -------
        ManimColorInternal
            Internal color representation
        """
        if len(hex) == 6:
            hex += "00"
        tmp = int(hex, 16)
        return np.asarray(
            (
                ((tmp >> 24) & 0xFF) / 255,
                ((tmp >> 16) & 0xFF) / 255,
                ((tmp >> 8) & 0xFF) / 255,
                alpha,
            ),
            dtype=ManimColorDType,
        )

    @staticmethod
    def _internal_from_int_rgb(
        rgb: RGB_Tuple_Int, alpha: float = 1.0
    ) -> ManimColorInternal:
        """Internal function for converting a rgb tuple of integers into the internal representation of a ManimColor.

        *For internal use only*

        Parameters
        ----------
        rgb : RGB_Tuple_Int
            integer rgb tuple to be parsed
        alpha : float, optional
            optional alpha value, by default 1.0

        Returns
        -------
        ManimColorInternal
            Internal color representation

        """
        value: np.ndarray = np.asarray(rgb, dtype=ManimColorDType).copy() / 255
        value.resize(4, refcheck=False)
        value[3] = alpha
        return value

    @staticmethod
    def _internal_from_rgb(
        rgb: RGB_Tuple_Float, alpha: float = 1.0
    ) -> ManimColorInternal:
        """Internal function for converting a rgb tuple of floats into the internal representation of a ManimColor.

        *For internal use only*

        Parameters
        ----------
        rgb : RGB_Tuple_Float
            float rgb tuple to be parsed

        alpha : float, optional
            optional alpha value, by default 1.0

        Returns
        -------
        ManimColorInternal
            Internal color representation
        """
        value: np.ndarray = np.asarray(rgb, dtype=ManimColorDType).copy()
        value.resize(4, refcheck=False)
        value[3] = alpha
        return value

    @staticmethod
    def _internal_from_int_rgba(rgba: RGBA_Tuple_Int) -> ManimColorInternal:
        """Internal function for converting a rgba tuple of integers into the internal representation of a ManimColor.

        *For internal use only*

        Parameters
        ----------
        rgba : RGBA_Tuple_Int
            int rgba tuple to be parsed

        Returns
        -------
        ManimColorInternal
            Internal color representation
        """
        return np.asarray(rgba, dtype=ManimColorDType) / 255

    @staticmethod
    def _internal_from_rgba(rgba: RGBA_Tuple_Float) -> ManimColorInternal:
        """Internal function for converting a rgba tuple of floats into the internal representation of a ManimColor.

        *For internal use only*

        Parameters
        ----------
        rgba : RGBA_Tuple_Float
            int rgba tuple to be parsed

        Returns
        -------
        ManimColorInternal
            Internal color representation
        """
        return np.asarray(rgba, dtype=ManimColorDType)

    @staticmethod
    def _internal_from_string(name: str) -> ManimColorInternal:
        """Internal function for converting a string into the internal representation of a ManimColor.
        This is not used for hex strings, please refer to :meth:`_internal_from_hex` for this functionality.

        *For internal use only*

        Parameters
        ----------
        name : str
            The color name to be parsed into a color. Refer to the different color Modules in the documentation Page to
            find the corresponding Color names.

        Returns
        -------
        ManimColorInternal
            Internal color representation

        Raises
        ------
        ValueError
            Raises a ValueError if the color name is not present with manim
        """
        from . import _all_color_dict

        upper_name = name.upper()

        if upper_name in _all_color_dict:
            return _all_color_dict[upper_name]._internal_value
        else:
            raise ValueError(f"Color {name} not found")

    def to_integer(self) -> int:
        """Converts the current ManimColor into an integer

        Returns
        -------
        int
            integer representation of the color

        .. warning::
            This will return only the rgb part of the color
        """
        return int.from_bytes(
            (self._internal_value[:3] * 255).astype(int).tobytes(), "big"
        )

    def to_rgb(self) -> RGB_Array_Float:
        """Converts the current ManimColor into a rgb array of floats

        Returns
        -------
        RGB_Array_Float
            rgb array with 3 elements of type float
        """
        return self._internal_value[:3]

    def to_int_rgb(self) -> RGB_Array_Int:
        """Converts the current ManimColor into a rgb array of int

        Returns
        -------
        RGB_Array_Int
            rgb array with 3 elements of type int
        """
        return (self._internal_value[:3] * 255).astype(int)

    def to_rgba(self) -> RGBA_Array_Float:
        """Converts the current ManimColor into a rgba array of floats

        Returns
        -------
        RGBA_Array_Float
            rgba array with 4 elements of type float
        """
        return self._internal_value

    def to_int_rgba(self) -> RGBA_Array_Int:
        """Converts the current ManimColor into a rgba array of int


        Returns
        -------
        RGBA_Array_Int
            rgba array with 4 elements of type int
        """
        return (self._internal_value * 255).astype(int)

    def to_rgba_with_alpha(self, alpha: float) -> RGBA_Array_Float:
        """Converts the current ManimColor into a rgba array of float as :meth:`to_rgba` but you can change the alpha
        value.

        Parameters
        ----------
        alpha : float
            alpha value to be used in the return value

        Returns
        -------
        RGBA_Array_Float
            rgba array with 4 elements of type float
        """
        return np.fromiter((*self._internal_value[:3], alpha), dtype=ManimColorDType)

    def to_int_rgba_with_alpha(self, alpha: float) -> RGBA_Array_Int:
        """Converts the current ManimColor into a rgba array of integers as :meth:`to_int_rgba` but you can change the alpha
        value.

        Parameters
        ----------
        alpha : float
            alpha value to be used for the return value. (Will automatically be scaled from 0-1 to 0-255 so just pass 0-1)

        Returns
        -------
        RGBA_Array_Int
            rgba array with 4 elements of type int
        """
        tmp = self._internal_value * 255
        tmp[3] = alpha * 255
        return tmp.astype(int)

    def to_hex(self, with_alpha: bool = False) -> str:
        """Converts the manim color to a hexadecimal representation of the color

        Parameters
        ----------
        with_alpha : bool, optional
            Changes the result from 6 to 8 values where the last 2 nibbles represent the alpha value of 0-255,
            by default False

        Returns
        -------
        str
            A hex string starting with a # with either 6 or 8 nibbles depending on your input, by default 6 i.e #XXXXXX
        """
        tmp = (
            f"#{int(self._internal_value[0] * 255):02X}"
            f"{int(self._internal_value[1] * 255):02X}"
            f"{int(self._internal_value[2] * 255):02X}"
        )
        if with_alpha:
            tmp += f"{int(self._internal_value[3] * 255):02X}"
        return tmp

    def to_hsv(self) -> HSV_Array_Float:
        """Converts the Manim Color to HSV array.

        .. note::
           Be careful this returns an array in the form `[h, s, v]` where the elements are floats.
           This might be confusing because rgb can also be an array of floats so you might want to annotate the usage
           of this function in your code by typing the variables with :class:`HSV_Array_Float` in order to differentiate
           between rgb arrays and hsv arrays

        Returns
        -------
        HSV_Array_Float
            A hsv array containing 3 elements of type float ranging from 0 to 1
        """
        return colorsys.rgb_to_hsv(*self.to_rgb())

    def invert(self, with_alpha=False) -> ManimColor:
        """Returns an linearly inverted version of the color (no inplace changes)

        Parameters
        ----------
        with_alpha : bool, optional
            if true the alpha value will be inverted too, by default False

            .. note::
                This can result in unintended behavior where objects are not displayed because their alpha
                value is suddenly 0 or very low. Please keep that in mind when setting this to true

        Returns
        -------
        ManimColor
            The linearly inverted ManimColor
        """
        return ManimColor(1.0 - self._internal_value, with_alpha)

    def interpolate(self, other: ManimColor, alpha: float) -> ManimColor:
        """Interpolates between the current and the given ManimColor an returns the interpolated color

        Parameters
        ----------
        other : ManimColor
            The other ManimColor to be used for interpolation
        alpha : float
            A point on the line in rgba colorspace connecting the two colors i.e. the interpolation point

            0 corresponds to the current ManimColor and 1 corresponds to the other ManimColor

        Returns
        -------
        ManimColor
            The interpolated ManimColor
        """
        return ManimColor(
            self._internal_value * (1 - alpha) + other._internal_value * alpha
        )

    @classmethod
    def from_rgb(
        cls,
        rgb: RGB_Array_Float | RGB_Tuple_Float | RGB_Array_Int | RGB_Tuple_Int,
        alpha: float = 1.0,
    ) -> Self:
        """Creates a ManimColor from an RGB Array. Automagically decides which type it is int/float

        .. warning::
            Please make sure that your elements are not floats if you want integers. A 5.0 will result in the input
            being interpreted as if it was a float rgb array with the value 5.0 and not the integer 5


        Parameters
        ----------
        rgb : RGB_Array_Float | RGB_Tuple_Float | RGB_Array_Int | RGB_Tuple_Int
            Any 3 Element Iterable
        alpha : float, optional
            alpha value to be used in the color, by default 1.0

        Returns
        -------
        ManimColor
            Returns the ManimColor object
        """
        return cls(rgb, alpha)

    @classmethod
    def from_rgba(
        cls, rgba: RGBA_Array_Float | RGBA_Tuple_Float | RGBA_Array_Int | RGBA_Tuple_Int
    ) -> Self:
        """Creates a ManimColor from an RGBA Array. Automagically decides which type it is int/float

        .. warning::
            Please make sure that your elements are not floats if you want integers. A 5.0 will result in the input
            being interpreted as if it was a float rgb array with the value 5.0 and not the integer 5

        Parameters
        ----------
        rgba : RGBA_Array_Float | RGBA_Tuple_Float | RGBA_Array_Int | RGBA_Tuple_Int
            Any 4 Element Iterable

        Returns
        -------
        ManimColor
            Returns the ManimColor object
        """
        return cls(rgba)

    @classmethod
    def from_hex(cls, hex: str, alpha: float = 1.0) -> Self:
        """Creates a Manim Color from a hex string, prefixes allowed # and 0x

        Parameters
        ----------
        hex : str
            The hex string to be converted (currently only supports 6 nibbles)
        alpha : float, optional
            alpha value to be used for the hex string, by default 1.0

        Returns
        -------
        ManimColor
            The ManimColor represented by the hex string
        """
        return cls(hex, alpha)

    @classmethod
    def from_hsv(
        cls, hsv: HSV_Array_Float | HSV_Tuple_Float, alpha: float = 1.0
    ) -> Self:
        """Creates a ManimColor from an HSV Array

        Parameters
        ----------
        hsv : HSV_Array_Float | HSV_Tuple_Float
            Any 3 Element Iterable containing floats from 0-1
        alpha : float, optional
            the alpha value to be used, by default 1.0

        Returns
        -------
        ManimColor
            The ManimColor with the corresponding RGB values to the HSV
        """
        rgb = colorsys.hsv_to_rgb(*hsv)
        return cls(rgb, alpha)

    @overload
    @classmethod
    def parse(
        cls,
        color: ParsableManimColor | None,
        alpha: float = ...,
    ) -> Self: ...

    @overload
    @classmethod
    def parse(
        cls,
        color: Sequence[ParsableManimColor],
        alpha: float = ...,
    ) -> list[Self]: ...

    @classmethod
    def parse(
        cls,
        color: ParsableManimColor | list[ParsableManimColor] | None,
        alpha: float = 1.0,
    ) -> Self | list[Self]:
        """
        Handles the parsing of a list of colors or a single color.

        Parameters
        ----------
        color
            The color or list of colors to parse. Note that this function can not accept rgba tuples. It will assume that you mean list[ManimColor] and will return a list of ManimColors.
        alpha
            The alpha value to use if a single color is passed. or if a list of colors is passed to set the value of all colors.

        Returns
        -------
        ManimColor
            Either a list of colors or a singular color depending on the input
        """
        if isinstance(color, (list, tuple)):
            return [cls(c, alpha) for c in color]  # type: ignore
        return cls(color, alpha)  # type: ignore

    @staticmethod
    def gradient(colors: list[ManimColor], length: int):
        """This is not implemented by now refer to :func:`color_gradient` for a working implementation for now"""
        # TODO: implement proper gradient, research good implementation for this or look at 3b1b implementation
        raise NotImplementedError

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self.to_hex()}')"

    def __str__(self) -> str:
        return f"{self.to_hex()}"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ManimColor):
            raise TypeError(
                f"Cannot compare {self.__class__.__name__} with {other.__class__.__name__}"
            )
        return np.allclose(self._internal_value, other._internal_value)

    def __add__(self, other: ManimColor) -> ManimColor:
        return ManimColor(self._internal_value + other._internal_value)

    def __sub__(self, other: ManimColor) -> ManimColor:
        return ManimColor(self._internal_value - other._internal_value)

    def __mul__(self, other: ManimColor) -> ManimColor:
        return ManimColor(self._internal_value * other._internal_value)

    def __truediv__(self, other: ManimColor) -> ManimColor:
        return ManimColor(self._internal_value / other._internal_value)

    def __floordiv__(self, other: ManimColor) -> ManimColor:
        return ManimColor(self._internal_value // other._internal_value)

    def __mod__(self, other: ManimColor) -> ManimColor:
        return ManimColor(self._internal_value % other._internal_value)

    def __pow__(self, other: ManimColor) -> ManimColor:
        return ManimColor(self._internal_value**other._internal_value)

    def __and__(self, other: ManimColor) -> ManimColor:
        return ManimColor(self.to_integer() & other.to_integer())

    def __or__(self, other: ManimColor) -> ManimColor:
        return ManimColor(self.to_integer() | other.to_integer())

    def __xor__(self, other: ManimColor) -> ManimColor:
        return ManimColor(self.to_integer() ^ other.to_integer())


ParsableManimColor: TypeAlias = Union[
    ManimColor,
    int,
    str,
    RGB_Tuple_Int,
    RGB_Tuple_Float,
    RGBA_Tuple_Int,
    RGBA_Tuple_Float,
    RGB_Array_Int,
    RGB_Array_Float,
    RGBA_Array_Int,
    RGBA_Array_Float,
]
"""`ParsableManimColor` represents all the types which can be parsed
to a color in Manim.
"""


ManimColorT = TypeVar("ManimColorT", bound=ManimColor)


def color_to_rgb(color: ParsableManimColor) -> RGB_Array_Float:
    """Helper function for use in functional style programming.
    Refer to :meth:`to_rgb` in :class:`ManimColor`.

    Parameters
    ----------
    color : ParsableManimColor
        A color

    Returns
    -------
    RGB_Array_Float
        the corresponding rgb array
    """
    return ManimColor(color).to_rgb()


def color_to_rgba(color: ParsableManimColor, alpha: float = 1) -> RGBA_Array_Float:
    """Helper function for use in functional style programming refer to :meth:`to_rgba_with_alpha` in :class:`ManimColor`

    Parameters
    ----------
    color : ParsableManimColor
        A color
    alpha : float, optional
        alpha value to be used in the color, by default 1

    Returns
    -------
    RGBA_Array_Float
        the corresponding rgba array
    """
    return ManimColor(color).to_rgba_with_alpha(alpha)


def color_to_int_rgb(color: ParsableManimColor) -> RGB_Array_Int:
    """Helper function for use in functional style programming refer to :meth:`to_int_rgb` in :class:`ManimColor`

    Parameters
    ----------
    color : ParsableManimColor
        A color

    Returns
    -------
    RGB_Array_Int
        the corresponding int rgb array
    """
    return ManimColor(color).to_int_rgb()


def color_to_int_rgba(color: ParsableManimColor, alpha: float = 1.0) -> RGBA_Array_Int:
    """Helper function for use in functional style programming refer to :meth:`to_int_rgba_with_alpha` in :class:`ManimColor`

    Parameters
    ----------
    color : ParsableManimColor
        A color
    alpha : float, optional
        alpha value to be used in the color, by default 1.0

    Returns
    -------
    RGBA_Array_Int
        the corresponding int rgba array
    """
    return ManimColor(color).to_int_rgba_with_alpha(alpha)


def rgb_to_color(
    rgb: RGB_Array_Float | RGB_Tuple_Float | RGB_Array_Int | RGB_Tuple_Int,
) -> ManimColor:
    """Helper function for use in functional style programming refer to :meth:`from_rgb` in :class:`ManimColor`

    Parameters
    ----------
    rgb : RGB_Array_Float | RGB_Tuple_Float
        A 3 element iterable

    Returns
    -------
    ManimColor
        A ManimColor with the corresponding value
    """
    return ManimColor.from_rgb(rgb)


def rgba_to_color(
    rgba: RGBA_Array_Float | RGBA_Tuple_Float | RGBA_Array_Int | RGBA_Tuple_Int,
) -> ManimColor:
    """Helper function for use in functional style programming refer to :meth:`from_rgba` in :class:`ManimColor`

    Parameters
    ----------
    rgba : RGBA_Array_Float | RGBA_Tuple_Float
        A 4 element iterable

    Returns
    -------
    ManimColor
        A ManimColor with the corresponding value
    """
    return ManimColor.from_rgba(rgba)


def rgb_to_hex(
    rgb: RGB_Array_Float | RGB_Tuple_Float | RGB_Array_Int | RGB_Tuple_Int,
) -> str:
    """Helper function for use in functional style programming refer to :meth:`from_rgb` in :class:`ManimColor`

    Parameters
    ----------
    rgb : RGB_Array_Float | RGB_Tuple_Float
        A 3 element iterable

    Returns
    -------
    str
        A hex representation of the color, refer to :meth:`to_hex` in :class:`ManimColor`
    """
    return ManimColor.from_rgb(rgb).to_hex()


def hex_to_rgb(hex_code: str) -> RGB_Array_Float:
    """Helper function for use in functional style programming refer to :meth:`to_hex` in :class:`ManimColor`

    Parameters
    ----------
    hex_code : str
        A hex string representing a color

    Returns
    -------
    RGB_Array_Float
        RGB array representing the color
    """
    return ManimColor(hex_code).to_rgb()


def invert_color(color: ManimColorT) -> ManimColorT:
    """Helper function for use in functional style programming refer to :meth:`invert` in :class:`ManimColor`

    Parameters
    ----------
    color : ManimColor
        A ManimColor

    Returns
    -------
    ManimColor
        The linearly inverted ManimColor
    """
    return color.invert()


def interpolate_arrays(
    arr1: npt.NDArray[Any], arr2: npt.NDArray[Any], alpha: float
) -> np.ndarray:
    """Helper function used in Manim to fade between two objects smoothly

    Parameters
    ----------
    arr1 : npt.NDArray[Any]
        The first array of colors
    arr2 : npt.NDArray[Any]
        The second array of colors
    alpha : float
        The alpha value corresponding to the interpolation point between the two inputs

    Returns
    -------
    np.ndarray
        The interpolated value of the to arrays
    """
    return (1 - alpha) * arr1 + alpha * arr2


def color_gradient(
    reference_colors: Sequence[ParsableManimColor],
    length_of_output: int,
) -> list[ManimColor] | ManimColor:
    """Creates a list of colors interpolated between the input array of colors with a specific number of colors

    Parameters
    ----------
    reference_colors : Sequence[ParsableManimColor]
        The colors to be interpolated between or spread apart
    length_of_output : int
        The number of colors that the output should have, ideally more than the input

    Returns
    -------
    list[ManimColor] | ManimColor
        A list of ManimColor's which has the interpolated colors
    """
    if length_of_output == 0:
        return ManimColor(reference_colors[0])
    if len(reference_colors) == 1:
        return [ManimColor(reference_colors[0])] * length_of_output
    rgbs = [color_to_rgb(color) for color in reference_colors]
    alphas = np.linspace(0, (len(rgbs) - 1), length_of_output)
    floors = alphas.astype("int")
    alphas_mod1 = alphas % 1
    # End edge case
    alphas_mod1[-1] = 1
    floors[-1] = len(rgbs) - 2
    return [
        rgb_to_color((rgbs[i] * (1 - alpha)) + (rgbs[i + 1] * alpha))
        for i, alpha in zip(floors, alphas_mod1)
    ]


def interpolate_color(
    color1: ManimColorT, color2: ManimColor, alpha: float
) -> ManimColorT:
    """Standalone function to interpolate two ManimColors and get the result refer to :meth:`interpolate` in :class:`ManimColor`

    Parameters
    ----------
    color1 : ManimColor
        First ManimColor
    color2 : ManimColor
        Second ManimColor
    alpha : float
        The alpha value determining the point of interpolation between the colors

    Returns
    -------
    ManimColor
        The interpolated ManimColor
    """
    return color1.interpolate(color2, alpha)


def average_color(*colors: ParsableManimColor) -> ManimColor:
    """Determines the Average color of the given parameters

    Returns
    -------
    ManimColor
        The average color of the input
    """
    rgbs = np.array([color_to_rgb(color) for color in colors])
    mean_rgb = np.apply_along_axis(np.mean, 0, rgbs)
    return rgb_to_color(mean_rgb)


def random_bright_color() -> ManimColor:
    """Returns you a random bright color

    .. warning::
        This operation is very expensive please keep in mind the performance loss.

    Returns
    -------
    ManimColor
        A bright ManimColor
    """
    curr_rgb = color_to_rgb(random_color())
    new_rgb = interpolate_arrays(curr_rgb, np.ones(len(curr_rgb)), 0.5)
    return ManimColor(new_rgb)


def random_color() -> ManimColor:
    """Return you a random ManimColor

    .. warning::
        This operation is very expensive please keep in mind the performance loss.

    Returns
    -------
    ManimColor
        _description_
    """
    import manim.utils.color.manim_colors as manim_colors

    return random.choice(manim_colors._all_manim_colors)


def get_shaded_rgb(
    rgb: npt.NDArray[Any],
    point: npt.NDArray[Any],
    unit_normal_vect: npt.NDArray[Any],
    light_source: npt.NDArray[Any],
) -> RGBA_Array_Float:
    to_sun = normalize(light_source - point)
    factor = 0.5 * np.dot(unit_normal_vect, to_sun) ** 3
    if factor < 0:
        factor *= 0.5
    result = rgb + factor
    return result


__all__ = [
    "ManimColor",
    "ManimColorDType",
    "ParsableManimColor",
    "color_to_rgb",
    "color_to_rgba",
    "color_to_int_rgb",
    "color_to_int_rgba",
    "rgb_to_color",
    "rgba_to_color",
    "rgb_to_hex",
    "hex_to_rgb",
    "invert_color",
    "interpolate_arrays",
    "color_gradient",
    "interpolate_color",
    "average_color",
    "random_bright_color",
    "random_color",
    "get_shaded_rgb",
]



"""Manim's default subcommand, render.

Manim's render subcommand is accessed in the command-line interface via
``manim``, but can be more explicitly accessed with ``manim render``. Here you
can specify options, and arguments for the render command.

"""

from __future__ import annotations

import http.client
import json
import sys
import urllib.error
import urllib.request
from pathlib import Path
from typing import cast

import cloup

from ... import __version__, config, console, error_console, logger
from ..._config import tempconfig
from ...constants import EPILOG, RendererType
from ...utils.module_ops import scene_classes_from_file
from .ease_of_access_options import ease_of_access_options
from .global_options import global_options
from .output_options import output_options
from .render_options import render_options

__all__ = ["render"]


@cloup.command(
    context_settings=None,
    no_args_is_help=True,
    epilog=EPILOG,
)
@cloup.argument("file", type=Path, required=True)
@cloup.argument("scene_names", required=False, nargs=-1)
@global_options
@output_options
@render_options  # type: ignore
@ease_of_access_options
def render(
    **args,
):
    """Render SCENE(S) from the input FILE.

    FILE is the file path of the script or a config file.

    SCENES is an optional list of scenes in the file.
    """

    if args["save_as_gif"]:
        logger.warning("--save_as_gif is deprecated, please use --format=gif instead!")
        args["format"] = "gif"

    if args["save_pngs"]:
        logger.warning("--save_pngs is deprecated, please use --format=png instead!")
        args["format"] = "png"

    if args["show_in_file_browser"]:
        logger.warning(
            "The short form of show_in_file_browser is deprecated and will be moved to support --format.",
        )

    class ClickArgs:
        def __init__(self, args):
            for name in args:
                setattr(self, name, args[name])

        def _get_kwargs(self):
            return list(self.__dict__.items())

        def __eq__(self, other):
            if not isinstance(other, ClickArgs):
                return NotImplemented
            return vars(self) == vars(other)

        def __contains__(self, key):
            return key in self.__dict__

        def __repr__(self):
            return str(self.__dict__)

    click_args = ClickArgs(args)
    if args["jupyter"]:
        return click_args

    config.digest_args(click_args)
    file = Path(config.input_file)
    if config.renderer == RendererType.OPENGL:
        from manim.renderer.opengl_renderer import OpenGLRenderer

        try:
            renderer = OpenGLRenderer()
            keep_running = True
            while keep_running:
                for SceneClass in scene_classes_from_file(file):
                    with tempconfig({}):
                        scene = SceneClass(renderer)
                        rerun = scene.render()
                    if rerun or config["write_all"]:
                        renderer.num_plays = 0
                        continue
                    else:
                        keep_running = False
                        break
                if config["write_all"]:
                    keep_running = False

        except Exception:
            error_console.print_exception()
            sys.exit(1)
    else:
        for SceneClass in scene_classes_from_file(file):
            try:
                with tempconfig({}):
                    scene = SceneClass()
                    scene.render()
            except Exception:
                error_console.print_exception()
                sys.exit(1)

    if config.notify_outdated_version:
        manim_info_url = "https://pypi.org/pypi/manim/json"
        warn_prompt = "Cannot check if latest release of manim is installed"

        try:
            with urllib.request.urlopen(
                urllib.request.Request(manim_info_url),
                timeout=10,
            ) as response:
                response = cast(http.client.HTTPResponse, response)
                json_data = json.loads(response.read())
        except urllib.error.HTTPError:
            logger.debug("HTTP Error: %s", warn_prompt)
        except urllib.error.URLError:
            logger.debug("URL Error: %s", warn_prompt)
        except json.JSONDecodeError:
            logger.debug(
                "Error while decoding JSON from %r: %s", manim_info_url, warn_prompt
            )
        except Exception:
            logger.debug("Something went wrong: %s", warn_prompt)
        else:
            stable = json_data["info"]["version"]
            if stable != __version__:
                console.print(
                    f"You are using manim version [red]v{__version__}[/red], but version [green]v{stable}[/green] is available.",
                )
                console.print(
                    "You should consider upgrading via [yellow]pip install -U manim[/yellow]",
                )

    return args



"""Utility functions related to Bézier curves."""

from __future__ import annotations

from manim.typing import (
    BezierPoints,
    ColVector,
    MatrixMN,
    Point3D,
    Point3D_Array,
    PointDType,
    QuadraticBezierPoints,
    QuadraticBezierPoints_Array,
)

__all__ = [
    "bezier",
    "partial_bezier_points",
    "partial_quadratic_bezier_points",
    "interpolate",
    "integer_interpolate",
    "mid",
    "inverse_interpolate",
    "match_interpolate",
    "get_smooth_handle_points",
    "get_smooth_cubic_bezier_handle_points",
    "diag_to_matrix",
    "is_closed",
    "proportions_along_bezier_curve_for_point",
    "point_lies_on_bezier",
]


from functools import reduce
from typing import Any, Callable, Sequence, overload

import numpy as np
import numpy.typing as npt
from scipy import linalg

from ..utils.simple_functions import choose
from ..utils.space_ops import cross2d, find_intersection


def bezier(
    points: Sequence[Point3D] | Point3D_Array,
) -> Callable[[float], Point3D]:
    """Classic implementation of a bezier curve.

    Parameters
    ----------
    points
        points defining the desired bezier curve.

    Returns
    -------
        function describing the bezier curve.
        You can pass a t value between 0 and 1 to get the corresponding point on the curve.
    """
    n = len(points) - 1
    # Cubic Bezier curve
    if n == 3:
        return lambda t: np.asarray(
            (1 - t) ** 3 * points[0]
            + 3 * t * (1 - t) ** 2 * points[1]
            + 3 * (1 - t) * t**2 * points[2]
            + t**3 * points[3],
            dtype=PointDType,
        )
    # Quadratic Bezier curve
    if n == 2:
        return lambda t: np.asarray(
            (1 - t) ** 2 * points[0] + 2 * t * (1 - t) * points[1] + t**2 * points[2],
            dtype=PointDType,
        )

    return lambda t: np.asarray(
        np.asarray(
            [
                (((1 - t) ** (n - k)) * (t**k) * choose(n, k) * point)
                for k, point in enumerate(points)
            ],
            dtype=PointDType,
        ).sum(axis=0)
    )


# !TODO: This function has still a weird implementation with the overlapping points
def partial_bezier_points(points: BezierPoints, a: float, b: float) -> BezierPoints:
    """Given an array of points which define bezier curve, and two numbers 0<=a<b<=1, return an array of the same size,
    which describes the portion of the original bezier curve on the interval [a, b].

    This algorithm is pretty nifty, and pretty dense.

    Parameters
    ----------
    points
        set of points defining the bezier curve.
    a
        lower bound of the desired partial bezier curve.
    b
        upper bound of the desired partial bezier curve.

    Returns
    -------
    np.ndarray
        Set of points defining the partial bezier curve.
    """
    _len = len(points)
    if a == 1:
        return np.asarray([points[-1]] * _len, dtype=PointDType)

    a_to_1 = np.asarray(
        [bezier(points[i:])(a) for i in range(_len)],
        dtype=PointDType,
    )
    end_prop = (b - a) / (1.0 - a)
    return np.asarray(
        [bezier(a_to_1[: i + 1])(end_prop) for i in range(_len)],
        dtype=PointDType,
    )


# Shortened version of partial_bezier_points just for quadratics,
# since this is called a fair amount
def partial_quadratic_bezier_points(
    points: QuadraticBezierPoints, a: float, b: float
) -> QuadraticBezierPoints:
    if a == 1:
        return np.asarray(3 * [points[-1]])

    def curve(t: float) -> Point3D:
        return np.asarray(
            points[0] * (1 - t) * (1 - t)
            + 2 * points[1] * t * (1 - t)
            + points[2] * t * t
        )

    # bezier(points)
    h0 = curve(a) if a > 0 else points[0]
    h2 = curve(b) if b < 1 else points[2]
    h1_prime = (1 - a) * points[1] + a * points[2]
    end_prop = (b - a) / (1.0 - a)
    h1 = (1 - end_prop) * h0 + end_prop * h1_prime
    return np.asarray((h0, h1, h2))


def split_quadratic_bezier(points: QuadraticBezierPoints, t: float) -> BezierPoints:
    """Split a quadratic Bézier curve at argument ``t`` into two quadratic curves.

    Parameters
    ----------
    points
        The control points of the bezier curve
        has shape ``[a1, h1, b1]``

    t
        The ``t``-value at which to split the Bézier curve

    Returns
    -------
        The two Bézier curves as a list of tuples,
        has the shape ``[a1, h1, b1], [a2, h2, b2]``
    """
    a1, h1, a2 = points
    s1 = interpolate(a1, h1, t)
    s2 = interpolate(h1, a2, t)
    p = interpolate(s1, s2, t)

    return np.array((a1, s1, p, p, s2, a2))


def subdivide_quadratic_bezier(points: QuadraticBezierPoints, n: int) -> BezierPoints:
    """Subdivide a quadratic Bézier curve into ``n`` subcurves which have the same shape.

    The points at which the curve is split are located at the
    arguments :math:`t = i/n` for :math:`i = 1, ..., n-1`.

    Parameters
    ----------
    points
        The control points of the Bézier curve in form ``[a1, h1, b1]``

    n
        The number of curves to subdivide the Bézier curve into

    Returns
    -------
        The new points for the Bézier curve in the form ``[a1, h1, b1, a2, h2, b2, ...]``

    .. image:: /_static/bezier_subdivision_example.png

    """
    beziers = np.empty((n, 3, 3))
    current = points
    for j in range(0, n):
        i = n - j
        tmp = split_quadratic_bezier(current, 1 / i)
        beziers[j] = tmp[:3]
        current = tmp[3:]
    return beziers.reshape(-1, 3)


def quadratic_bezier_remap(
    triplets: QuadraticBezierPoints_Array, new_number_of_curves: int
) -> QuadraticBezierPoints_Array:
    """Remaps the number of curves to a higher amount by splitting bezier curves

    Parameters
    ----------
    triplets
        The triplets of the quadratic bezier curves to be remapped shape(n, 3, 3)

    new_number_of_curves
        The number of curves that the output will contain. This needs to be higher than the current number.

    Returns
    -------
        The new triplets for the quadratic bezier curves.
    """
    difference = new_number_of_curves - len(triplets)
    if difference <= 0:
        return triplets
    new_triplets = np.zeros((new_number_of_curves, 3, 3))
    idx = 0
    for triplet in triplets:
        if difference > 0:
            tmp_noc = int(np.ceil(difference / len(triplets))) + 1
            tmp = subdivide_quadratic_bezier(triplet, tmp_noc).reshape(-1, 3, 3)
            for i in range(tmp_noc):
                new_triplets[idx + i] = tmp[i]
            difference -= tmp_noc - 1
            idx += tmp_noc
        else:
            new_triplets[idx] = triplet
            idx += 1
    return new_triplets

    """
    This is an alternate version of the function just for documentation purposes
    --------

    difference = new_number_of_curves - len(triplets)
    if difference <= 0:
        return triplets
    new_triplets = []
    for triplet in triplets:
        if difference > 0:
            tmp_noc = int(np.ceil(difference / len(triplets))) + 1
            tmp = subdivide_quadratic_bezier(triplet, tmp_noc).reshape(-1, 3, 3)
            for i in range(tmp_noc):
                new_triplets.append(tmp[i])
            difference -= tmp_noc - 1
        else:
            new_triplets.append(triplet)
    return new_triplets
    """


# Linear interpolation variants


@overload
def interpolate(start: float, end: float, alpha: float) -> float: ...


@overload
def interpolate(start: Point3D, end: Point3D, alpha: float) -> Point3D: ...


def interpolate(
    start: int | float | Point3D, end: int | float | Point3D, alpha: float | Point3D
) -> float | Point3D:
    return (1 - alpha) * start + alpha * end


def integer_interpolate(
    start: float,
    end: float,
    alpha: float,
) -> tuple[int, float]:
    """
    This is a variant of interpolate that returns an integer and the residual

    Parameters
    ----------
    start
        The start of the range
    end
        The end of the range
    alpha
        a float between 0 and 1.

    Returns
    -------
    tuple[int, float]
        This returns an integer between start and end (inclusive) representing
        appropriate interpolation between them, along with a
        "residue" representing a new proportion between the
        returned integer and the next one of the
        list.

    Example
    -------

    .. code-block:: pycon

        >>> integer, residue = integer_interpolate(start=0, end=10, alpha=0.46)
        >>> np.allclose((integer, residue), (4, 0.6))
        True
    """
    if alpha >= 1:
        return (int(end - 1), 1.0)
    if alpha <= 0:
        return (int(start), 0)
    value = int(interpolate(start, end, alpha))
    residue = ((end - start) * alpha) % 1
    return (value, residue)


@overload
def mid(start: float, end: float) -> float: ...


@overload
def mid(start: Point3D, end: Point3D) -> Point3D: ...


def mid(start: float | Point3D, end: float | Point3D) -> float | Point3D:
    """Returns the midpoint between two values.

    Parameters
    ----------
    start
        The first value
    end
        The second value

    Returns
    -------
        The midpoint between the two values
    """
    return (start + end) / 2.0


@overload
def inverse_interpolate(start: float, end: float, value: float) -> float: ...


@overload
def inverse_interpolate(start: float, end: float, value: Point3D) -> Point3D: ...


@overload
def inverse_interpolate(start: Point3D, end: Point3D, value: Point3D) -> Point3D: ...


def inverse_interpolate(
    start: float | Point3D, end: float | Point3D, value: float | Point3D
) -> float | Point3D:
    """Perform inverse interpolation to determine the alpha
    values that would produce the specified ``value``
    given the ``start`` and ``end`` values or points.

    Parameters
    ----------
    start
        The start value or point of the interpolation.
    end
        The end value or point of the interpolation.
    value
        The value or point for which the alpha value
        should be determined.

    Returns
    -------
        The alpha values producing the given input
        when interpolating between ``start`` and ``end``.

    Example
    -------

    .. code-block:: pycon

        >>> inverse_interpolate(start=2, end=6, value=4)
        0.5

        >>> start = np.array([1, 2, 1])
        >>> end = np.array([7, 8, 11])
        >>> value = np.array([4, 5, 5])
        >>> inverse_interpolate(start, end, value)
        array([0.5, 0.5, 0.4])
    """
    return np.true_divide(value - start, end - start)


@overload
def match_interpolate(
    new_start: float,
    new_end: float,
    old_start: float,
    old_end: float,
    old_value: float,
) -> float: ...


@overload
def match_interpolate(
    new_start: float,
    new_end: float,
    old_start: float,
    old_end: float,
    old_value: Point3D,
) -> Point3D: ...


def match_interpolate(
    new_start: float,
    new_end: float,
    old_start: float,
    old_end: float,
    old_value: float | Point3D,
) -> float | Point3D:
    """Interpolate a value from an old range to a new range.

    Parameters
    ----------
    new_start
        The start of the new range.
    new_end
        The end of the new range.
    old_start
        The start of the old range.
    old_end
        The end of the old range.
    old_value
        The value within the old range whose corresponding
        value in the new range (with the same alpha value)
        is desired.

    Returns
    -------
        The interpolated value within the new range.

    Examples
    --------
    >>> match_interpolate(0, 100, 10, 20, 15)
    50.0
    """
    old_alpha = inverse_interpolate(old_start, old_end, old_value)
    return interpolate(
        new_start,
        new_end,
        old_alpha,  # type: ignore
    )


def get_smooth_cubic_bezier_handle_points(
    points: Point3D_Array,
) -> tuple[BezierPoints, BezierPoints]:
    points = np.asarray(points)
    num_handles = len(points) - 1
    dim = points.shape[1]
    if num_handles < 1:
        return np.zeros((0, dim)), np.zeros((0, dim))
    # Must solve 2*num_handles equations to get the handles.
    # l and u are the number of lower an upper diagonal rows
    # in the matrix to solve.
    l, u = 2, 1
    # diag is a representation of the matrix in diagonal form
    # See https://www.particleincell.com/2012/bezier-splines/
    # for how to arrive at these equations
    diag: MatrixMN = np.zeros((l + u + 1, 2 * num_handles))
    diag[0, 1::2] = -1
    diag[0, 2::2] = 1
    diag[1, 0::2] = 2
    diag[1, 1::2] = 1
    diag[2, 1:-2:2] = -2
    diag[3, 0:-3:2] = 1
    # last
    diag[2, -2] = -1
    diag[1, -1] = 2
    # This is the b as in Ax = b, where we are solving for x,
    # and A is represented using diag.  However, think of entries
    # to x and b as being points in space, not numbers
    b: Point3D_Array = np.zeros((2 * num_handles, dim))
    b[1::2] = 2 * points[1:]
    b[0] = points[0]
    b[-1] = points[-1]

    def solve_func(b: ColVector) -> ColVector | MatrixMN:
        return linalg.solve_banded((l, u), diag, b)  # type: ignore

    use_closed_solve_function = is_closed(points)
    if use_closed_solve_function:
        # Get equations to relate first and last points
        matrix = diag_to_matrix((l, u), diag)
        # last row handles second derivative
        matrix[-1, [0, 1, -2, -1]] = [2, -1, 1, -2]
        # first row handles first derivative
        matrix[0, :] = np.zeros(matrix.shape[1])
        matrix[0, [0, -1]] = [1, 1]
        b[0] = 2 * points[0]
        b[-1] = np.zeros(dim)

        def closed_curve_solve_func(b: ColVector) -> ColVector | MatrixMN:
            return linalg.solve(matrix, b)  # type: ignore

    handle_pairs = np.zeros((2 * num_handles, dim))
    for i in range(dim):
        if use_closed_solve_function:
            handle_pairs[:, i] = closed_curve_solve_func(b[:, i])
        else:
            handle_pairs[:, i] = solve_func(b[:, i])
    return handle_pairs[0::2], handle_pairs[1::2]


def get_smooth_handle_points(
    points: BezierPoints,
) -> tuple[BezierPoints, BezierPoints]:
    """Given some anchors (points), compute handles so the resulting bezier curve is smooth.

    Parameters
    ----------
    points
        Anchors.

    Returns
    -------
    typing.Tuple[np.ndarray, np.ndarray]
        Computed handles.
    """
    # NOTE points here are anchors.
    points = np.asarray(points)
    num_handles = len(points) - 1
    dim = points.shape[1]
    if num_handles < 1:
        return np.zeros((0, dim)), np.zeros((0, dim))
    # Must solve 2*num_handles equations to get the handles.
    # l and u are the number of lower an upper diagonal rows
    # in the matrix to solve.
    l, u = 2, 1
    # diag is a representation of the matrix in diagonal form
    # See https://www.particleincell.com/2012/bezier-splines/
    # for how to arrive at these equations
    diag: MatrixMN = np.zeros((l + u + 1, 2 * num_handles))
    diag[0, 1::2] = -1
    diag[0, 2::2] = 1
    diag[1, 0::2] = 2
    diag[1, 1::2] = 1
    diag[2, 1:-2:2] = -2
    diag[3, 0:-3:2] = 1
    # last
    diag[2, -2] = -1
    diag[1, -1] = 2
    # This is the b as in Ax = b, where we are solving for x,
    # and A is represented using diag.  However, think of entries
    # to x and b as being points in space, not numbers
    b = np.zeros((2 * num_handles, dim))
    b[1::2] = 2 * points[1:]
    b[0] = points[0]
    b[-1] = points[-1]

    def solve_func(b: ColVector) -> ColVector | MatrixMN:
        return linalg.solve_banded((l, u), diag, b)  # type: ignore

    use_closed_solve_function = is_closed(points)
    if use_closed_solve_function:
        # Get equations to relate first and last points
        matrix = diag_to_matrix((l, u), diag)
        # last row handles second derivative
        matrix[-1, [0, 1, -2, -1]] = [2, -1, 1, -2]
        # first row handles first derivative
        matrix[0, :] = np.zeros(matrix.shape[1])
        matrix[0, [0, -1]] = [1, 1]
        b[0] = 2 * points[0]
        b[-1] = np.zeros(dim)

        def closed_curve_solve_func(b: ColVector) -> ColVector | MatrixMN:
            return linalg.solve(matrix, b)  # type: ignore

    handle_pairs = np.zeros((2 * num_handles, dim))
    for i in range(dim):
        if use_closed_solve_function:
            handle_pairs[:, i] = closed_curve_solve_func(b[:, i])
        else:
            handle_pairs[:, i] = solve_func(b[:, i])
    return handle_pairs[0::2], handle_pairs[1::2]


def diag_to_matrix(
    l_and_u: tuple[int, int], diag: npt.NDArray[Any]
) -> npt.NDArray[Any]:
    """
    Converts array whose rows represent diagonal
    entries of a matrix into the matrix itself.
    See scipy.linalg.solve_banded
    """
    l, u = l_and_u
    dim = diag.shape[1]
    matrix = np.zeros((dim, dim))
    for i in range(l + u + 1):
        np.fill_diagonal(
            matrix[max(0, i - u) :, max(0, u - i) :],
            diag[i, max(0, u - i) :],
        )
    return matrix


# Given 4 control points for a cubic bezier curve (or arrays of such)
# return control points for 2 quadratics (or 2n quadratics) approximating them.
def get_quadratic_approximation_of_cubic(
    a0: Point3D, h0: Point3D, h1: Point3D, a1: Point3D
) -> BezierPoints:
    a0 = np.array(a0, ndmin=2)
    h0 = np.array(h0, ndmin=2)
    h1 = np.array(h1, ndmin=2)
    a1 = np.array(a1, ndmin=2)
    # Tangent vectors at the start and end.
    T0 = h0 - a0
    T1 = a1 - h1

    # Search for inflection points.  If none are found, use the
    # midpoint as a cut point.
    # Based on http://www.caffeineowl.com/graphics/2d/vectorial/cubic-inflexion.html
    has_infl = np.ones(len(a0), dtype=bool)

    p = h0 - a0
    q = h1 - 2 * h0 + a0
    r = a1 - 3 * h1 + 3 * h0 - a0

    a = cross2d(q, r)
    b = cross2d(p, r)
    c = cross2d(p, q)

    disc = b * b - 4 * a * c
    has_infl &= disc > 0
    sqrt_disc = np.sqrt(np.abs(disc))
    settings = np.seterr(all="ignore")
    ti_bounds = []
    for sgn in [-1, +1]:
        ti = (-b + sgn * sqrt_disc) / (2 * a)
        ti[a == 0] = (-c / b)[a == 0]
        ti[(a == 0) & (b == 0)] = 0
        ti_bounds.append(ti)
    ti_min, ti_max = ti_bounds
    np.seterr(**settings)
    ti_min_in_range = has_infl & (0 < ti_min) & (ti_min < 1)
    ti_max_in_range = has_infl & (0 < ti_max) & (ti_max < 1)

    # Choose a value of t which starts at 0.5,
    # but is updated to one of the inflection points
    # if they lie between 0 and 1

    t_mid = 0.5 * np.ones(len(a0))
    t_mid[ti_min_in_range] = ti_min[ti_min_in_range]
    t_mid[ti_max_in_range] = ti_max[ti_max_in_range]

    m, n = a0.shape
    t_mid = t_mid.repeat(n).reshape((m, n))

    # Compute bezier point and tangent at the chosen value of t (these are vectorized)
    mid = bezier([a0, h0, h1, a1])(t_mid)  # type: ignore
    Tm = bezier([h0 - a0, h1 - h0, a1 - h1])(t_mid)  # type: ignore

    # Intersection between tangent lines at end points
    # and tangent in the middle
    i0 = find_intersection(a0, T0, mid, Tm)
    i1 = find_intersection(a1, T1, mid, Tm)

    m, n = np.shape(a0)
    result = np.zeros((6 * m, n))
    result[0::6] = a0
    result[1::6] = i0
    result[2::6] = mid
    result[3::6] = mid
    result[4::6] = i1
    result[5::6] = a1
    return result


def is_closed(points: Point3D_Array) -> bool:
    return np.allclose(points[0], points[-1])  # type: ignore


def proportions_along_bezier_curve_for_point(
    point: Point3D,
    control_points: BezierPoints,
    round_to: float = 1e-6,
) -> npt.NDArray[Any]:
    """Obtains the proportion along the bezier curve corresponding to a given point
    given the bezier curve's control points.

    The bezier polynomial is constructed using the coordinates of the given point
    as well as the bezier curve's control points. On solving the polynomial for each dimension,
    if there are roots common to every dimension, those roots give the proportion along the
    curve the point is at. If there are no real roots, the point does not lie on the curve.

    Parameters
    ----------
    point
        The Cartesian Coordinates of the point whose parameter
        should be obtained.
    control_points
        The Cartesian Coordinates of the ordered control
        points of the bezier curve on which the point may
        or may not lie.
    round_to
        A float whose number of decimal places all values
        such as coordinates of points will be rounded.

    Returns
    -------
        np.ndarray[float]
            List containing possible parameters (the proportions along the bezier curve)
            for the given point on the given bezier curve.
            This usually only contains one or zero elements, but if the
            point is, say, at the beginning/end of a closed loop, may return
            a list with more than 1 value, corresponding to the beginning and
            end etc. of the loop.

    Raises
    ------
    :class:`ValueError`
        When ``point`` and the control points have different shapes.
    """
    # Method taken from
    # http://polymathprogrammer.com/2012/04/03/does-point-lie-on-bezier-curve/

    if not all(np.shape(point) == np.shape(c_p) for c_p in control_points):
        raise ValueError(
            f"Point {point} and Control Points {control_points} have different shapes.",
        )

    control_points = np.array(control_points)
    n = len(control_points) - 1

    roots = []
    for dim, coord in enumerate(point):
        control_coords = control_points[:, dim]
        terms = []
        for term_power in range(n, -1, -1):
            outercoeff = choose(n, term_power)
            term = []
            sign = 1
            for subterm_num in range(term_power, -1, -1):
                innercoeff = choose(term_power, subterm_num) * sign
                subterm = innercoeff * control_coords[subterm_num]
                if term_power == 0:
                    subterm -= coord
                term.append(subterm)
                sign *= -1
            terms.append(outercoeff * sum(np.array(term)))
        if all(term == 0 for term in terms):
            # Then both Bezier curve and Point lie on the same plane.
            # Roots will be none, but in this specific instance, we don't need to consider that.
            continue
        bezier_polynom = np.polynomial.Polynomial(terms[::-1])
        polynom_roots = bezier_polynom.roots()  # type: ignore
        if len(polynom_roots) > 0:
            polynom_roots = np.around(polynom_roots, int(np.log10(1 / round_to)))
        roots.append(polynom_roots)

    roots = [[root for root in rootlist if root.imag == 0] for rootlist in roots]
    # Get common roots
    # arg-type: ignore
    roots = reduce(np.intersect1d, roots)  # type: ignore
    result = np.asarray([r.real for r in roots if 0 <= r.real <= 1])
    return result


def point_lies_on_bezier(
    point: Point3D,
    control_points: BezierPoints,
    round_to: float = 1e-6,
) -> bool:
    """Checks if a given point lies on the bezier curves with the given control points.

    This is done by solving the bezier polynomial with the point as the constant term; if
    any real roots exist, the point lies on the bezier curve.

    Parameters
    ----------
    point
        The Cartesian Coordinates of the point to check.
    control_points
        The Cartesian Coordinates of the ordered control
        points of the bezier curve on which the point may
        or may not lie.
    round_to
        A float whose number of decimal places all values
        such as coordinates of points will be rounded.

    Returns
    -------
    bool
        Whether the point lies on the curve.
    """

    roots = proportions_along_bezier_curve_for_point(point, control_points, round_to)

    return len(roots) > 0



"""Auxiliary module for the checkhealth subcommand, contains
the actual check implementations."""

from __future__ import annotations

import os
import shutil
import subprocess
from typing import Callable

from ..._config import config

__all__ = ["HEALTH_CHECKS"]

HEALTH_CHECKS = []


def healthcheck(
    description: str,
    recommendation: str,
    skip_on_failed: list[Callable | str] | None = None,
    post_fail_fix_hook: Callable | None = None,
):
    """Decorator used for declaring health checks.

    This decorator attaches some data to a function,
    which is then added to a list containing all checks.

    Parameters
    ----------
    description
        A brief description of this check, displayed when
        the checkhealth subcommand is run.
    recommendation
        Help text which is displayed in case the check fails.
    skip_on_failed
        A list of check functions which, if they fail, cause
        the current check to be skipped.
    post_fail_fix_hook
        A function that is supposed to (interactively) help
        to fix the detected problem, if possible. This is
        only called upon explicit confirmation of the user.

    Returns
    -------
    A check function, as required by the checkhealth subcommand.
    """
    if skip_on_failed is None:
        skip_on_failed = []
    skip_on_failed = [
        skip.__name__ if callable(skip) else skip for skip in skip_on_failed
    ]

    def decorator(func):
        func.description = description
        func.recommendation = recommendation
        func.skip_on_failed = skip_on_failed
        func.post_fail_fix_hook = post_fail_fix_hook
        HEALTH_CHECKS.append(func)
        return func

    return decorator


@healthcheck(
    description="Checking whether manim is on your PATH",
    recommendation=(
        "The command <manim> is currently not on your system's PATH.\n\n"
        "You can work around this by calling the manim module directly "
        "via <python -m manim> instead of just <manim>.\n\n"
        "To fix the PATH issue properly: "
        "Usually, the Python package installer pip issues a warning "
        "during the installation which contains more information. "
        "Consider reinstalling manim via <pip uninstall manim> "
        "followed by <pip install manim> to see the warning again, "
        "then consult the internet on how to modify your system's "
        "PATH variable."
    ),
)
def is_manim_on_path():
    path_to_manim = shutil.which("manim")
    return path_to_manim is not None


@healthcheck(
    description="Checking whether the executable belongs to manim",
    recommendation=(
        "The command <manim> does not belong to your installed version "
        "of this library, it likely belongs to manimgl / manimlib.\n\n"
        "Run manim via <python -m manim> or via <manimce>, or uninstall "
        "and reinstall manim via <pip install --upgrade "
        "--force-reinstall manim> to fix this."
    ),
    skip_on_failed=[is_manim_on_path],
)
def is_manim_executable_associated_to_this_library():
    path_to_manim = shutil.which("manim")
    with open(path_to_manim, "rb") as f:
        manim_exec = f.read()

    # first condition below corresponds to the executable being
    # some sort of python script. second condition happens when
    # the executable is actually a Windows batch file.
    return b"manim.__main__" in manim_exec or b'"%~dp0\\manim"' in manim_exec


@healthcheck(
    description="Checking whether ffmpeg is available",
    recommendation=(
        "Manim does not work without ffmpeg. Please follow our "
        "installation instructions "
        "at https://docs.manim.community/en/stable/installation.html "
        "to download ffmpeg. Then, either ...\n\n"
        "(a) ... make the ffmpeg executable available to your system's PATH,\n"
        "(b) or, alternatively, use <manim cfg write --open> to create a "
        "custom configuration and set the ffmpeg_executable variable to the "
        "full absolute path to the ffmpeg executable."
    ),
)
def is_ffmpeg_available():
    path_to_ffmpeg = shutil.which(config.ffmpeg_executable)
    return path_to_ffmpeg is not None and os.access(path_to_ffmpeg, os.X_OK)


@healthcheck(
    description="Checking whether ffmpeg is working",
    recommendation=(
        "Your installed version of ffmpeg does not support x264 encoding, "
        "which manim requires. Please follow our installation instructions "
        "at https://docs.manim.community/en/stable/installation.html "
        "to download and install a newer version of ffmpeg."
    ),
    skip_on_failed=[is_ffmpeg_available],
)
def is_ffmpeg_working():
    ffmpeg_version = subprocess.run(
        [config.ffmpeg_executable, "-version"],
        stdout=subprocess.PIPE,
    ).stdout.decode()
    return (
        ffmpeg_version.startswith("ffmpeg version")
        and "--enable-libx264" in ffmpeg_version
    )


@healthcheck(
    description="Checking whether latex is available",
    recommendation=(
        "Manim cannot find <latex> on your system's PATH. "
        "You will not be able to use Tex and MathTex mobjects "
        "in your scenes.\n\n"
        "Consult our installation instructions "
        "at https://docs.manim.community/en/stable/installation.html "
        "or search the web for instructions on how to install a "
        "LaTeX distribution on your operating system."
    ),
)
def is_latex_available():
    path_to_latex = shutil.which("latex")
    return path_to_latex is not None and os.access(path_to_latex, os.X_OK)


@healthcheck(
    description="Checking whether dvisvgm is available",
    recommendation=(
        "Manim could find <latex>, but not <dvisvgm> on your system's "
        "PATH. Make sure your installed LaTeX distribution comes with "
        "dvisvgm and consider installing a larger distribution if it "
        "does not."
    ),
    skip_on_failed=[is_latex_available],
)
def is_dvisvgm_available():
    path_to_dvisvgm = shutil.which("dvisvgm")
    return path_to_dvisvgm is not None and os.access(path_to_dvisvgm, os.X_OK)



r"""Mobjects that inherit from lines and contain a label along the length."""

from __future__ import annotations

__all__ = ["LabeledLine", "LabeledArrow"]

from manim.constants import *
from manim.mobject.geometry.line import Arrow, Line
from manim.mobject.geometry.shape_matchers import (
    BackgroundRectangle,
    SurroundingRectangle,
)
from manim.mobject.text.tex_mobject import MathTex, Tex
from manim.mobject.text.text_mobject import Text
from manim.utils.color import WHITE, ManimColor, ParsableManimColor


class LabeledLine(Line):
    """Constructs a line containing a label box somewhere along its length.

    Parameters
    ----------
    label : str | Tex | MathTex | Text
        Label that will be displayed on the line.
    label_position : float | optional
        A ratio in the range [0-1] to indicate the position of the label with respect to the length of the line. Default value is 0.5.
    font_size : float | optional
        Control font size for the label. This parameter is only used when `label` is of type `str`.
    label_color: ParsableManimColor | optional
        The color of the label's text. This parameter is only used when `label` is of type `str`.
    label_frame : Bool | optional
        Add a `SurroundingRectangle` frame to the label box.
    frame_fill_color : ParsableManimColor | optional
        Background color to fill the label box. If no value is provided, the background color of the canvas will be used.
    frame_fill_opacity : float | optional
        Determine the opacity of the label box by passing a value in the range [0-1], where 0 indicates complete transparency and 1 means full opacity.

    .. seealso::
        :class:`LabeledArrow`

    Examples
    --------
    .. manim:: LabeledLineExample
        :save_last_frame:

        class LabeledLineExample(Scene):
            def construct(self):
                line = LabeledLine(
                    label          = '0.5',
                    label_position = 0.8,
                    font_size      = 20,
                    label_color    = WHITE,
                    label_frame    = True,

                    start=LEFT+DOWN,
                    end=RIGHT+UP)


                line.set_length(line.get_length() * 2)
                self.add(line)
    """

    def __init__(
        self,
        label: str | Tex | MathTex | Text,
        label_position: float = 0.5,
        font_size: float = DEFAULT_FONT_SIZE,
        label_color: ParsableManimColor = WHITE,
        label_frame: bool = True,
        frame_fill_color: ParsableManimColor = None,
        frame_fill_opacity: float = 1,
        *args,
        **kwargs,
    ) -> None:
        label_color = ManimColor(label_color)
        frame_fill_color = ManimColor(frame_fill_color)
        if isinstance(label, str):
            from manim import MathTex

            rendered_label = MathTex(label, color=label_color, font_size=font_size)
        else:
            rendered_label = label

        super().__init__(*args, **kwargs)

        # calculating the vector for the label position
        line_start, line_end = self.get_start_and_end()
        new_vec = (line_end - line_start) * label_position
        label_coords = line_start + new_vec

        # rendered_label.move_to(self.get_vector() * label_position)
        rendered_label.move_to(label_coords)

        box = BackgroundRectangle(
            rendered_label,
            buff=0.05,
            color=frame_fill_color,
            fill_opacity=frame_fill_opacity,
            stroke_width=0.5,
        )
        self.add(box)

        if label_frame:
            box_frame = SurroundingRectangle(
                rendered_label, buff=0.05, color=label_color, stroke_width=0.5
            )

            self.add(box_frame)

        self.add(rendered_label)


class LabeledArrow(LabeledLine, Arrow):
    """Constructs an arrow containing a label box somewhere along its length.
    This class inherits its label properties from `LabeledLine`, so the main parameters controlling it are the same.

    Parameters
    ----------
    label : str | Tex | MathTex | Text
        Label that will be displayed on the line.
    label_position : float | optional
        A ratio in the range [0-1] to indicate the position of the label with respect to the length of the line. Default value is 0.5.
    font_size : float | optional
        Control font size for the label. This parameter is only used when `label` is of type `str`.
    label_color: ParsableManimColor | optional
        The color of the label's text. This parameter is only used when `label` is of type `str`.
    label_frame : Bool | optional
        Add a `SurroundingRectangle` frame to the label box.
    frame_fill_color : ParsableManimColor | optional
        Background color to fill the label box. If no value is provided, the background color of the canvas will be used.
    frame_fill_opacity : float | optional
        Determine the opacity of the label box by passing a value in the range [0-1], where 0 indicates complete transparency and 1 means full opacity.


    .. seealso::
        :class:`LabeledLine`

    Examples
    --------
    .. manim:: LabeledArrowExample
        :save_last_frame:

        class LabeledArrowExample(Scene):
            def construct(self):
                l_arrow = LabeledArrow("0.5", start=LEFT*3, end=RIGHT*3 + UP*2, label_position=0.5)

                self.add(l_arrow)
    """

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)



"""A directive for documenting colors in Manim."""

from __future__ import annotations

import inspect
from typing import TYPE_CHECKING

from docutils import nodes
from docutils.parsers.rst import Directive

from manim import ManimColor

if TYPE_CHECKING:
    from sphinx.application import Sphinx

__all__ = ["ManimColorModuleDocumenter"]


def setup(app: Sphinx) -> None:
    app.add_directive("automanimcolormodule", ManimColorModuleDocumenter)


class ManimColorModuleDocumenter(Directive):
    objtype = "automanimcolormodule"
    required_arguments = 1
    has_content = True

    def add_directive_header(self, sig: str) -> None:
        super().add_directive_header(sig)

    def run(self) -> list[nodes.Element]:
        module_name = self.arguments[0]
        try:
            import importlib

            module = importlib.import_module(module_name)
        except ImportError:
            return [
                nodes.error(
                    None,
                    nodes.paragraph(text="Failed to import module '%s'" % module_name),
                )
            ]

        # Number of Colors displayed in one row
        num_color_cols = 2
        table = nodes.table(align="center")

        tgroup = nodes.tgroup(cols=num_color_cols * 2)
        table += tgroup
        for _ in range(num_color_cols * 2):
            tgroup += nodes.colspec(colwidth=1)

        # Create header rows for the table
        thead = nodes.thead()
        row = nodes.row()
        for _ in range(num_color_cols):
            col1 = nodes.paragraph(text="Color Name")
            col2 = nodes.paragraph(text="RGB Hex Code")
            row += nodes.entry("", col1)
            row += nodes.entry("", col2)
        thead += row
        tgroup += thead

        color_elements = []
        for member_name, member_obj in inspect.getmembers(module):
            if isinstance(member_obj, ManimColor):
                r, g, b = member_obj.to_rgb()
                luminance = 0.2126 * r + 0.7152 * g + 0.0722 * b

                # Choose the font color based on the background luminance
                if luminance > 0.5:
                    font_color = "black"
                else:
                    font_color = "white"

                color_elements.append((member_name, member_obj.to_hex(), font_color))

        tbody = nodes.tbody()

        for base_i in range(0, len(color_elements), num_color_cols):
            row = nodes.row()
            for member_name, hex_code, font_color in color_elements[
                base_i : base_i + num_color_cols
            ]:
                col1 = nodes.literal(text=member_name)
                col2 = nodes.raw(
                    "",
                    f'<div style="background-color:{hex_code};padding: 0.25rem 0;border-radius:8px;margin: 0.5rem 0.2rem"><code style="color:{font_color};">{hex_code}</code></div>',
                    format="html",
                )
                row += nodes.entry("", col1)
                row += nodes.entry("", col2)
            tbody += row
        tgroup += tbody

        return [table]
